using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;
using CymaticLabs.Unity3D.Amqp;


/// <summary>
/// Uses the AMQP Client on a custom way for our application
/// </summary>
public class MessageHandler : MonoBehaviour
{
    public static MessageHandler Instance { get; private set; }

    // List of created exchange subscriptions
    List<AmqpExchangeSubscription> exSubscriptions;

    // The current list of exchanges
    AmqpExchange[] exchanges;

    #region Methods
    private void Awake()
    {
        Instance = this;
        // Supongo que la linea de abajo resetea las suscripciones cuando 
        // hace el awake pero no se como funciona
        //exSubscriptions = new List<AmqpExchangeSubscription>();
    }

    private void Start()
    {
        AmqpClient.Instance.OnConnected.AddListener(HandleConnected);
        AmqpClient.Instance.OnDisconnected.AddListener(HandleDisconnected);
        AmqpClient.Instance.OnReconnecting.AddListener(HandleReconnecting);
        AmqpClient.Instance.OnBlocked.AddListener(HandleBlocked);
        AmqpClient.Instance.OnSubscribedToExchange.AddListener(HandleExchangeSubscribed);
        AmqpClient.Instance.OnUnsubscribedFromExchange.AddListener(HandleExchangeUnsubscribed);
        
    }

    #region Connect

    /// <summary>
    /// Connects to the AMQP broker using the form's client connection values.
    /// </summary>
    public void Connect(string connectionName)
    {
        // Connect
        AmqpClient.Instance.Connection = connectionName;

        AmqpClient.Connect();
    }

    #endregion Connect

    #region Disconnect

    /// <summary>
    /// Disconnects the AMQP client.
    /// </summary>
    public void Disconnect()
    {
        // Disconnect
        AmqpClient.Disconnect();
    }

    #endregion Disconnect

    #region Subscribe

    /// <summary>
    /// Subscribes to the AMQP exchange subscription using the form's values.
    /// </summary>
    public void Subscribe(string exchangeName, string routingKey)
    {
        // Validate args
        var isValid = true;

        // Don't continue if values are invald
        if (string.IsNullOrEmpty(exchangeName)) return;

        var exchangeType = AmqpExchangeTypes.Topic;

        // Find this exchange and get its exchange type
        foreach (var exchange in exchanges)
        {
            if (exchange.Name == exchangeName)
            {
                exchangeType = exchange.Type;
                break;
            }
        }

        // Ensure this subscription doesn't already exist
        foreach (var sub in exSubscriptions)
        {
            if (sub.ExchangeName == exchangeName && sub.ExchangeType == exchangeType && sub.RoutingKey == routingKey)
            {
                return;
            }
        }

        // Create the new subscription
        //Tener cuidado porque el chabon escribio mal exchange al implementarlo.
        var subscription = new UnityAmqpExchangeSubscription(exchangeName, exchangeType, routingKey, null, AmqpClient.Instance.UnityEventDebugExhangeMessageHandler);

        // Subscribe on the client
        AmqpClient.Subscribe(subscription);
    }

    #endregion Subscribe

    #region Unsubscribe

    /// <summary>
    /// Unsubscribes from the AMQP exchange subscription using the form's values.
    /// </summary>
    public void Unsubscribe(string exchangeName, string routingKey)
    {
        // Don't continue if values are invald
        if (string.IsNullOrEmpty(exchangeName)) return;

        var exchangeType = AmqpExchangeTypes.Topic;

        // Find this exchange and get its exchange type
        foreach (var exchange in exchanges)
        {
            if (exchange.Name == exchangeName)
            {
                exchangeType = exchange.Type;
                break;
            }
        }

        // Ensure this subscription already exists
        var subs = exSubscriptions.ToArray();

        foreach (var sub in subs)
        {
            if (sub.ExchangeName == exchangeName && sub.ExchangeType == exchangeType && sub.RoutingKey == routingKey)
            {
                AmqpClient.Unsubscribe(sub);
                exSubscriptions.Remove(sub);
                return;
            }
        }
    }

    #endregion Unsubscribe

    #region Publish

    /// <summary>
    /// Publishes a message to the current exchange using the form's input values.
    /// </summary>
    public void Publish(string exchangeName, string routingKey, string message)
    {
        // Validate args
        // Don't continue if values are invald
        if (string.IsNullOrEmpty(exchangeName)) return;

        var exchangeType = AmqpExchangeTypes.Topic;
        // Publish the message
        AmqpClient.Publish(exchangeName, routingKey, message);
    }

    #endregion Publish

    #region Event Handlers

    // Handles a connection event
    void HandleConnected(AmqpClient client)
    {
        Debug.Log("New connection");
        // Query exchange list
        AmqpClient.GetExchangesAsync(FinishConnected);
    }

    // Finishes the connection event by receiving the async result of the exchange query and display the results in the drop down
    void FinishConnected(AmqpExchange[] exchangeList)
    {
        //No se que hace esto 
    }

    // Handles a disconnection event
    void HandleDisconnected(AmqpClient client)
    {

    }

    // Handles a reconnecting event
    void HandleReconnecting(AmqpClient client)
    {

    }

    // Handles a blocked event
    void HandleBlocked(AmqpClient client)
    {

    }

    // Handles exchange subscribes
    void HandleExchangeSubscribed(AmqpExchangeSubscription subscription)
    {
        // Add it to the local list
        exSubscriptions.Add(subscription);
    }

    // Handles exchange unsubscribes
    void HandleExchangeUnsubscribed(AmqpExchangeSubscription subscription)
    {
        // Add it to the local list
        exSubscriptions.Remove(subscription);
    }

    #endregion Event Handlers

    #endregion Methods
}
